@@ -20,6 +20,8 @@
 #include <sys/stat.h>
 #include <sys/uio.h>
 #include <gelf.h>
+#include <dlfcn.h>
+#include <link.h>
 
 /* This should be defined before #include "utils.h" */
 #define PR_FMT     "mcount"
@@ -34,13 +36,14 @@
 
 uint64_t mcount_threshold;  /* nsec */
 struct symtabs symtabs = {
-	.flags = SYMTAB_FL_DEMANGLE | SYMTAB_FL_ADJ_OFFSET,
+	.flags = SYMTAB_FL_DEMANGLE | SYMTAB_FL_ADJ_OFFSET |
+		 SYMTAB_FL_SKIP_NORMAL | SYMTAB_FL_SKIP_DYNAMIC,
 };
 int shmem_bufsize = SHMEM_BUFFER_SIZE;
 bool mcount_setup_done;
 bool mcount_finished;
 
-pthread_key_t mtd_key;
+pthread_key_t mtd_key = (pthread_key_t)-1;
 TLS struct mcount_thread_data mtd;
 
 static int pfd = -1;
@@ -72,7 +75,7 @@ int gettid(struct mcount_thread_data *mtdp)
 
 const char *session_name(void)
 {
-	static char session[16 + 1];
+	static char session[SESSION_ID_LEN + 1];
 	static uint64_t session_id;
 	int fd;
 
@@ -86,7 +89,8 @@ const char *session_name(void)
 
 		close(fd);
 
-		snprintf(session, sizeof(session), "%016"PRIx64, session_id);
+		snprintf(session, sizeof(session), "%0*"PRIx64,
+			 SESSION_ID_LEN, session_id);
 	}
 	return session;
 }
@@ -142,54 +146,102 @@ static void send_session_msg(struct mcount_thread_data *mtdp, const char *sess_i
 		pr_err("write tid info failed");
 }
 
+static void send_dlopen_msg(struct mcount_thread_data *mtdp, const char *sess_id,
+			    uint64_t timestamp,  uint64_t base_addr,
+			    const char *libname)
+{
+	struct ftrace_msg_dlopen dlop = {
+		.task = {
+			.time = timestamp,
+			.pid = getpid(),
+			.tid = gettid(mtdp),
+		},
+		.base_addr = base_addr,
+		.namelen = strlen(libname),
+	};
+	struct ftrace_msg msg = {
+		.magic = FTRACE_MSG_MAGIC,
+		.type = FTRACE_MSG_DLOPEN,
+		.len = sizeof(dlop) + dlop.namelen,
+	};
+	struct iovec iov[3] = {
+		{ .iov_base = &msg, .iov_len = sizeof(msg), },
+		{ .iov_base = &dlop, .iov_len = sizeof(dlop), },
+		{ .iov_base = (void *)libname, .iov_len = dlop.namelen, },
+	};
+	int len = sizeof(msg) + msg.len;
+
+	if (pfd < 0)
+		return;
+
+	memcpy(dlop.sid, sess_id, sizeof(dlop.sid));
+
+	if (writev(pfd, iov, 3) != len)
+		pr_err("write tid info failed");
+}
+
 /* to be used by pthread_create_key() */
 static void mtd_dtor(void *arg)
 {
 	struct mcount_thread_data *mtdp = arg;
 
+	/* this thread is done, do not enter anymore */
+	mtdp->recursion_guard = true;
+
 	free(mtdp->rstack);
+	mtdp->rstack = NULL;
+
 #ifndef DISABLE_MCOUNT_FILTER
 	free(mtdp->argbuf);
+	mtdp->argbuf = NULL;
 #endif
 	shmem_finish(mtdp);
 }
 
 static void mcount_init_file(void)
 {
-	/* This is for the case of library-only tracing */
-	if (!mcount_setup_done)
-		__monstartup(0, ~0);
-
-	if (pthread_key_create(&mtd_key, mtd_dtor))
-		pr_err("cannot create shmem key");
-
 	send_session_msg(&mtd, session_name());
 }
 
-void mcount_prepare(void)
+struct mcount_thread_data * mcount_prepare(void)
 {
 	static pthread_once_t once_control = PTHREAD_ONCE_INIT;
-	struct ftrace_msg_task tmsg = {
-		.pid = getpid(),
-		.tid = gettid(&mtd),
-	};
+	struct mcount_thread_data *mtdp = &mtd;
+	struct ftrace_msg_task tmsg;
+
+	/*
+	 * If an executable implements its own malloc(),
+	 * following recursion could occur
+	 *
+	 * mcount_entry -> mcount_prepare -> xmalloc -> mcount_entry -> ...
+	 */
+	if (mtdp->recursion_guard)
+		return NULL;
+
+	mtdp->recursion_guard = true;
+	compiler_barrier();
 
 #ifndef DISABLE_MCOUNT_FILTER
-	mtd.filter.depth  = mcount_depth;
-	mtd.enable_cached = mcount_enabled;
-	mtd.argbuf = xmalloc(mcount_rstack_max * ARGBUF_SIZE);
+	mtdp->filter.depth  = mcount_depth;
+	mtdp->filter.time   = mcount_threshold;
+	mtdp->enable_cached = mcount_enabled;
+	mtdp->argbuf = xmalloc(mcount_rstack_max * ARGBUF_SIZE);
 #endif
-	mtd.rstack = xmalloc(mcount_rstack_max * sizeof(*mtd.rstack));
+	mtdp->rstack = xmalloc(mcount_rstack_max * sizeof(*mtd.rstack));
 
 	pthread_once(&once_control, mcount_init_file);
-	prepare_shmem_buffer(&mtd);
+	prepare_shmem_buffer(mtdp);
 
-	pthread_setspecific(mtd_key, &mtd);
+	pthread_setspecific(mtd_key, mtdp);
 
 	/* time should be get after session message sent */
+	tmsg.pid = getpid(),
+	tmsg.tid = gettid(mtdp),
 	tmsg.time = mcount_gettime();
 
 	ftrace_send_message(FTRACE_MSG_TID, &tmsg, sizeof(tmsg));
+
+	return mtdp;
 }
 
 bool mcount_check_rstack(struct mcount_thread_data *mtdp)
@@ -217,14 +269,15 @@ enum filter_result mcount_entry_filter_check(struct mcount_thread_data *mtdp,
 	if (mcount_check_rstack(mtdp))
 		return FILTER_RSTACK;
 
-	/* save original depth to restore at exit time */
+	/* save original depth and time to restore at exit time */
 	mtdp->filter.saved_depth = mtdp->filter.depth;
+	mtdp->filter.saved_time  = mtdp->filter.time;
 
 	/* already filtered by notrace option */
 	if (mtdp->filter.out_count > 0)
 		return FILTER_OUT;
 
-	ftrace_match_filter(&mcount_triggers, child, tr);
+	uftrace_match_filter(child, &mcount_triggers, tr);
 
 	pr_dbg3(" tr->flags: %lx, filter mode, count: [%d] %d/%d\n",
 		tr->flags, mcount_filter_mode, mtdp->filter.in_count,
@@ -246,7 +299,8 @@ enum filter_result mcount_entry_filter_check(struct mcount_thread_data *mtdp,
 			return FILTER_OUT;
 	}
 
-#define FLAGS_TO_CHECK  (TRIGGER_FL_DEPTH | TRIGGER_FL_TRACE_ON | TRIGGER_FL_TRACE_OFF)
+#define FLAGS_TO_CHECK  (TRIGGER_FL_DEPTH | TRIGGER_FL_TRACE_ON |	\
+			 TRIGGER_FL_TRACE_OFF | TRIGGER_FL_TIME_FILTER)
 
 	if (tr->flags & FLAGS_TO_CHECK) {
 		if (tr->flags & TRIGGER_FL_DEPTH)
@@ -257,6 +311,9 @@ enum filter_result mcount_entry_filter_check(struct mcount_thread_data *mtdp,
 
 		if (tr->flags & TRIGGER_FL_TRACE_OFF)
 			mcount_enabled = false;
+
+		if (tr->flags & TRIGGER_FL_TIME_FILTER)
+			mtdp->filter.time = tr->time;
 	}
 
 #undef FLAGS_TO_CHECK
@@ -282,6 +339,7 @@ void mcount_entry_filter_record(struct mcount_thread_data *mtdp,
 		rstack->flags |= MCOUNT_FL_NORECORD;
 
 	rstack->filter_depth = mtdp->filter.saved_depth;
+	rstack->filter_time  = mtdp->filter.saved_time;
 
 #define FLAGS_TO_CHECK  (TRIGGER_FL_FILTER | TRIGGER_FL_RETVAL | TRIGGER_FL_TRACE)
 
@@ -329,7 +387,7 @@ void mcount_entry_filter_record(struct mcount_thread_data *mtdp,
 		}
 
 		if (tr->flags & TRIGGER_FL_RECOVER) {
-			mcount_restore();
+			mcount_rstack_restore();
 			*rstack->parent_loc = (unsigned long) mcount_return;
 			rstack->flags |= MCOUNT_FL_RECOVER;
 		}
@@ -341,6 +399,8 @@ void mcount_exit_filter_record(struct mcount_thread_data *mtdp,
 			       struct mcount_ret_stack *rstack,
 			       long *retval)
 {
+	uint64_t time_filter = mtdp->filter.time;
+
 	pr_dbg3("<%d> exit  %lx\n", mtdp->idx, rstack->child_ip);
 
 #define FLAGS_TO_CHECK  (MCOUNT_FL_FILTERED | MCOUNT_FL_NOTRACE | MCOUNT_FL_RECOVER)
@@ -352,12 +412,13 @@ void mcount_exit_filter_record(struct mcount_thread_data *mtdp,
 			mtdp->filter.out_count--;
 
 		if (rstack->flags & MCOUNT_FL_RECOVER)
-			mcount_reset();
+			mcount_rstack_reset();
 	}
 
 #undef FLAGS_TO_CHECK
 
 	mtdp->filter.depth = rstack->filter_depth;
+	mtdp->filter.time  = rstack->filter_time;
 
 	if (!(rstack->flags & MCOUNT_FL_NORECORD)) {
 		if (mtdp->record_idx > 0)
@@ -366,7 +427,7 @@ void mcount_exit_filter_record(struct mcount_thread_data *mtdp,
 		if (!(rstack->flags & MCOUNT_FL_RETVAL))
 			retval = NULL;
 
-		if (rstack->end_time - rstack->start_time > mcount_threshold ||
+		if (rstack->end_time - rstack->start_time > time_filter ||
 		    rstack->flags & (MCOUNT_FL_WRITTEN | MCOUNT_FL_TRACE)) {
 			if (!mcount_enabled)
 				return;
@@ -428,23 +489,21 @@ int mcount_entry(unsigned long *parent_loc, unsigned long child,
 		.flags = 0,
 	};
 
-	/*
-	 * If an executable has its own malloc(), following recursion could occur
-	 *
-	 * mcount_entry -> mcount_prepare -> xmalloc -> mcount_entry -> ...
-	 */
 	if (unlikely(mcount_should_stop()))
 		return -1;
 
-	mtd.recursion_guard = true;
-
 	/* Access the mtd through TSD pointer to reduce TLS overhead */
 	mtdp = get_thread_data();
 	if (unlikely(check_thread_data(mtdp))) {
-		mcount_prepare();
+		mtdp = mcount_prepare();
+		if (mtdp == NULL)
+			return -1;
+	}
+	else {
+		if (unlikely(mtdp->recursion_guard))
+			return -1;
 
-		mtdp = get_thread_data();
-		assert(mtdp);
+		mtdp->recursion_guard = true;
 	}
 
 	filtered = mcount_entry_filter_check(mtdp, child, &tr);
@@ -529,19 +588,28 @@ static int cygprof_entry(unsigned long parent, unsigned long child)
 	if (unlikely(mcount_should_stop()))
 		return -1;
 
-	mtd.recursion_guard = true;
-
 	/* Access the mtd through TSD pointer to reduce TLS overhead */
 	mtdp = get_thread_data();
 	if (unlikely(check_thread_data(mtdp))) {
-		mcount_prepare();
+		mtdp = mcount_prepare();
+		if (mtdp == NULL)
+			return -1;
+	}
+	else {
+		if (unlikely(mtdp->recursion_guard))
+			return -1;
 
-		mtdp = get_thread_data();
-		assert(mtdp);
+		mtdp->recursion_guard = true;
 	}
 
 	filtered = mcount_entry_filter_check(mtdp, child, &tr);
 
+	/* 
+	 * recording arguments and return value is not supported.
+	 * also 'recover' trigger is only work for -pg entry.
+	 */
+	tr.flags &= ~(TRIGGER_FL_ARGUMENT | TRIGGER_FL_RETVAL | TRIGGER_FL_RECOVER);
+
 	rstack = &mtdp->rstack[mtdp->idx++];
 
 	/*
@@ -555,6 +623,7 @@ static int cygprof_entry(unsigned long parent, unsigned long child)
 
 	rstack->depth      = mtdp->record_idx;
 	rstack->dyn_idx    = MCOUNT_INVALID_DYNIDX;
+	rstack->parent_loc = &mtdp->cygprof_dummy;
 	rstack->parent_ip  = parent;
 	rstack->child_ip   = child;
 	rstack->end_time   = 0;
@@ -581,14 +650,17 @@ static void cygprof_exit(unsigned long parent, unsigned long child)
 	if (unlikely(mcount_should_stop()))
 		return;
 
-	mtd.recursion_guard = true;
-
 	mtdp = get_thread_data();
 	if (unlikely(check_thread_data(mtdp))) {
-		mcount_prepare();
+		mtdp = mcount_prepare();
+		if (mtdp == NULL)
+			return;
+	}
+	else {
+		if (unlikely(mtdp->recursion_guard))
+			return;
 
-		mtdp = get_thread_data();
-		assert(mtdp);
+		mtdp->recursion_guard = true;
 	}
 
 	/*
@@ -634,18 +706,23 @@ static void atfork_child_handler(void)
 
 	mtdp = get_thread_data();
 	if (unlikely(check_thread_data(mtdp))) {
-		mcount_prepare();
+		/* we need it even if in a recursion */
+		mtd.recursion_guard = false;
 
-		mtdp = get_thread_data();
-		assert(mtdp);
+		mtdp = mcount_prepare();
 	}
 
-	mtd.tid = 0;
+	/* flush tid cache */
+	mtdp->tid = 0;
 
-	clear_shmem_buffer(&mtd);
-	prepare_shmem_buffer(&mtd);
+	mtdp->recursion_guard = true;
+
+	clear_shmem_buffer(mtdp);
+	prepare_shmem_buffer(mtdp);
 
 	ftrace_send_message(FTRACE_MSG_FORK_END, &tmsg, sizeof(tmsg));
+
+	mtdp->recursion_guard = false;
 }
 
 static void build_debug_domain(char *dbg_domain_str)
@@ -671,10 +748,77 @@ static void build_debug_domain(char *dbg_domain_str)
 	}
 }
 
-/*
- * external interfaces
- */
-void __visible_default __monstartup(unsigned long low, unsigned long high)
+struct dlopen_base_data {
+	const char *libname;
+	unsigned long base_addr;
+};
+
+static const char *simple_basename(const char *pathname)
+{
+	const char *p = strrchr(pathname, '/');
+
+	return p ? p + 1 : pathname;
+}
+
+static int dlopen_base_callback(struct dl_phdr_info *info,
+				size_t size, void *arg)
+{
+	struct dlopen_base_data *data = arg;
+
+	if (!strstr(simple_basename(info->dlpi_name), data->libname))
+		return 0;
+
+	data->base_addr = info->dlpi_addr;
+	data->libname = info->dlpi_name; /* update to use full path */
+	return 0;
+}
+
+static void (*old_segfault_handler)(int);
+
+static void segfault_handler(int sig)
+{
+	mcount_rstack_restore();
+
+	signal(sig, old_segfault_handler);
+	raise(sig);
+}
+
+void mcount_rstack_restore(void)
+{
+	int idx;
+	struct mcount_thread_data *mtdp;
+
+	mtdp = get_thread_data();
+	if (unlikely(check_thread_data(mtdp)))
+		return;
+
+	for (idx = mtdp->idx - 1; idx >= 0; idx--)
+		*mtdp->rstack[idx].parent_loc = mtdp->rstack[idx].parent_ip;
+}
+
+void mcount_rstack_reset(void)
+{
+	int idx;
+	struct mcount_thread_data *mtdp;
+	struct mcount_ret_stack *rstack;
+
+	mtdp = get_thread_data();
+	if (unlikely(check_thread_data(mtdp)))
+		return;
+
+	for (idx = mtdp->idx - 1; idx >= 0; idx--) {
+		rstack = &mtdp->rstack[idx];
+
+		if (rstack->dyn_idx == MCOUNT_INVALID_DYNIDX)
+			*rstack->parent_loc = (unsigned long)mcount_return;
+		else
+			*rstack->parent_loc = (unsigned long)plthook_return;
+	}
+}
+
+static void mcount_hook_functions(void);
+
+static void mcount_startup(void)
 {
 	char *pipefd_str;
 	char *logfd_str;
@@ -684,6 +828,11 @@ void __visible_default __monstartup(unsigned long low, unsigned long high)
 	char *threshold_str;
 	char *color_str;
 	char *demangle_str;
+	char *filter_str;
+	char *trigger_str;
+	char *argument_str;
+	char *retval_str;
+	char *plthook_str;
 	char *dirname;
 	struct stat statbuf;
 	LIST_HEAD(modules);
@@ -696,6 +845,9 @@ void __visible_default __monstartup(unsigned long low, unsigned long high)
 	outfp = stdout;
 	logfp = stderr;
 
+	if (pthread_key_create(&mtd_key, mtd_dtor))
+		pr_err("cannot create mtd key");
+
 	pipefd_str = getenv("UFTRACE_PIPE");
 	logfd_str = getenv("UFTRACE_LOGFD");
 	debug_str = getenv("UFTRACE_DEBUG");
@@ -704,6 +856,11 @@ void __visible_default __monstartup(unsigned long low, unsigned long high)
 	color_str = getenv("UFTRACE_COLOR");
 	threshold_str = getenv("UFTRACE_THRESHOLD");
 	demangle_str = getenv("UFTRACE_DEMANGLE");
+	filter_str = getenv("UFTRACE_FILTER");
+	trigger_str = getenv("UFTRACE_TRIGGER");
+	argument_str = getenv("UFTRACE_ARGUMENT");
+	retval_str = getenv("UFTRACE_RETVAL");
+	plthook_str = getenv("UFTRACE_PLTHOOK");
 
 	if (logfd_str) {
 		int fd = strtol(logfd_str, NULL, 0);
@@ -715,6 +872,8 @@ void __visible_default __monstartup(unsigned long low, unsigned long high)
 		}
 	}
 
+	old_segfault_handler = signal(SIGSEGV, segfault_handler);
+
 	if (debug_str) {
 		debug = strtol(debug_str, NULL, 0);
 		build_debug_domain(getenv("UFTRACE_DEBUG_DOMAIN"));
@@ -745,29 +904,29 @@ void __visible_default __monstartup(unsigned long low, unsigned long high)
 	if (dirname == NULL)
 		dirname = UFTRACE_DIR_NAME;
 
+	if (filter_str || trigger_str || argument_str || retval_str)
+		symtabs.flags &= ~SYMTAB_FL_SKIP_NORMAL;
+	if (plthook_str)
+		symtabs.flags &= ~SYMTAB_FL_SKIP_DYNAMIC;
+
 	mcount_exename = read_exename();
 	record_proc_maps(dirname, session_name(), &symtabs);
+	set_kernel_base(dirname, session_name());
 	load_symtabs(&symtabs, NULL, mcount_exename);
 
 #ifndef DISABLE_MCOUNT_FILTER
-	ftrace_setup_filter_module(getenv("UFTRACE_FILTER"), &modules);
-	ftrace_setup_filter_module(getenv("UFTRACE_TRIGGER"), &modules);
-	ftrace_setup_filter_module(getenv("UFTRACE_ARGUMENT"), &modules);
-	ftrace_setup_filter_module(getenv("UFTRACE_RETVAL"), &modules);
+	ftrace_setup_filter_module(filter_str, &modules, mcount_exename);
+	ftrace_setup_filter_module(trigger_str, &modules, mcount_exename);
+	ftrace_setup_filter_module(argument_str, &modules, mcount_exename);
+	ftrace_setup_filter_module(retval_str, &modules, mcount_exename);
 
 	load_module_symtabs(&symtabs, &modules);
 
-	ftrace_setup_filter(getenv("UFTRACE_FILTER"), &symtabs, NULL,
-			    &mcount_triggers, &mcount_filter_mode);
-
-	ftrace_setup_trigger(getenv("UFTRACE_TRIGGER"), &symtabs, NULL,
-			     &mcount_triggers);
-
-	ftrace_setup_argument(getenv("UFTRACE_ARGUMENT"), &symtabs, NULL,
-			      &mcount_triggers);
-
-	ftrace_setup_retval(getenv("UFTRACE_RETVAL"), &symtabs, NULL,
-			      &mcount_triggers);
+	ftrace_setup_filter(filter_str, &symtabs, &mcount_triggers,
+			    &mcount_filter_mode);
+	ftrace_setup_trigger(trigger_str, &symtabs, &mcount_triggers);
+	ftrace_setup_argument(argument_str, &symtabs, &mcount_triggers);
+	ftrace_setup_retval(retval_str, &symtabs, &mcount_triggers);
 
 	if (getenv("UFTRACE_DEPTH"))
 		mcount_depth = strtol(getenv("UFTRACE_DEPTH"), NULL, 0);
@@ -782,28 +941,14 @@ void __visible_default __monstartup(unsigned long low, unsigned long high)
 	if (threshold_str)
 		mcount_threshold = strtoull(threshold_str, NULL, 0);
 
-	if (getenv("UFTRACE_PLTHOOK")) {
-		if (symtabs.loaded && symtabs.dsymtab.nr_sym == 0) {
+	if (plthook_str) {
+		if (symtabs.dsymtab.nr_sym == 0) {
 			pr_dbg("skip PLT hooking due to no dynamic symbols\n");
 			goto out;
 		}
 
 		setup_dynsym_indexes(&symtabs);
 
-#ifndef DISABLE_MCOUNT_FILTER
-		ftrace_setup_filter(getenv("UFTRACE_FILTER"), &symtabs, "PLT",
-				    &mcount_triggers, &mcount_filter_mode);
-
-		ftrace_setup_trigger(getenv("UFTRACE_TRIGGER"), &symtabs, "PLT",
-				    &mcount_triggers);
-
-		ftrace_setup_argument(getenv("UFTRACE_ARGUMENT"), &symtabs, "PLT",
-				      &mcount_triggers);
-
-		ftrace_setup_retval(getenv("UFTRACE_RETVAL"), &symtabs, "PLT",
-				      &mcount_triggers);
-#endif /* DISABLE_MCOUNT_FILTER */
-
 		if (hook_pltgot(mcount_exename, symtabs.maps->start) < 0)
 			pr_dbg("error when hooking plt: skipping...\n");
 		else
@@ -813,17 +958,20 @@ void __visible_default __monstartup(unsigned long low, unsigned long high)
 out:
 	pthread_atfork(atfork_prepare_handler, NULL, atfork_child_handler);
 
+	mcount_hook_functions();
+
 #ifndef DISABLE_MCOUNT_FILTER
 	ftrace_cleanup_filter_module(&modules);
 #endif /* DISABLE_MCOUNT_FILTER */
 
 	compiler_barrier();
+	pr_dbg("mcount setup done\n");
 
 	mcount_setup_done = true;
 	mtd.recursion_guard = false;
 }
 
-void __visible_default _mcleanup(void)
+static void mcount_cleanup(void)
 {
 	mcount_finish();
 	destroy_dynsym_indexes();
@@ -833,30 +981,159 @@ void __visible_default _mcleanup(void)
 #endif
 }
 
-void __visible_default mcount_restore(void)
+/*
+ * hooking functions
+ */
+static int (*real_backtrace)(void **buffer, int sz);
+static void (*real_cxa_throw)(void *exc, void *type, void *dest);
+static void (*real_cxa_end_catch)(void);
+static void * (*real_dlopen)(const char *filename, int flags);
+
+static void mcount_hook_functions(void)
+{
+	real_backtrace		= dlsym(RTLD_NEXT, "backtrace");
+	real_cxa_throw		= dlsym(RTLD_NEXT, "__cxa_throw");
+	real_cxa_end_catch	= dlsym(RTLD_NEXT, "__cxa_end_catch");
+	real_dlopen		= dlsym(RTLD_NEXT, "dlopen");
+}
+
+__visible_default int backtrace(void **buffer, int sz)
+{
+	int ret;
+
+	if (real_backtrace == NULL)
+		return 0;
+
+	mcount_rstack_restore();
+	ret = real_backtrace(buffer, sz);
+	mcount_rstack_reset();
+
+	return ret;
+}
+
+__visible_default void __cxa_throw(void *exception, void *type, void *dest)
 {
-	int idx;
 	struct mcount_thread_data *mtdp;
 
-	mtdp = get_thread_data();
-	if (unlikely(check_thread_data(mtdp)))
-		return;
+	/*
+	 * restore return addresses so that it can unwind stack frames
+	 * safely during the exception handling.
+	 * It pairs to __cxa_end_catch().
+	 */
+	mcount_rstack_restore();
 
-	for (idx = mtdp->idx - 1; idx >= 0; idx--)
-		*mtdp->rstack[idx].parent_loc = mtdp->rstack[idx].parent_ip;
+	mtdp = get_thread_data();
+	if (!check_thread_data(mtdp))
+		pr_dbg("exception thrown from [%d]\n", mtdp->idx);
+
+	real_cxa_throw(exception, type, dest);
+}
+
+__visible_default void __cxa_end_catch(void)
+{
+	struct mcount_thread_data *mtdp;
+	struct mcount_ret_stack *rstack;
+	unsigned long retaddr;
+
+	/* get frame address where exception handler returns */
+	retaddr = (unsigned long)__builtin_frame_address(0);
+
+	real_cxa_end_catch();
+
+	pr_dbg("exception returned at frame: %#lx\n", retaddr);
+
+	mcount_rstack_restore();
+
+	mtdp = get_thread_data();
+	if (!check_thread_data(mtdp)) {
+		int idx;
+
+		/* it needs to find how much stack frame was unwinded */
+		for (idx = mtdp->idx - 1; idx >= 0; idx--) {
+			rstack = &mtdp->rstack[idx];
+
+			pr_dbg2("[%d] parent at %p\n", idx, rstack->parent_loc);
+			if (rstack->parent_loc == &mtdp->cygprof_dummy)
+				break;
+
+			if ((unsigned long)rstack->parent_loc > retaddr)
+				break;
+
+			/* record unwinded functions */
+			if (!(rstack->flags & MCOUNT_FL_NORECORD))
+				rstack->end_time = mcount_gettime();
+
+			mcount_exit_filter_record(mtdp, rstack, NULL);
+		}
+
+		/* we're in ENTER state, so add 1 to the index */
+		mtdp->idx = idx + 1;
+		pr_dbg("[%d] exception returned\n", mtdp->idx);
+
+		mcount_rstack_reset();
+	}
+}
+
+__visible_default void * dlopen(const char *filename, int flags)
+{
+	struct mcount_thread_data *mtdp;
+	uint64_t timestamp = mcount_gettime();
+	void *ret = real_dlopen(filename, flags);
+	struct dlopen_base_data data = {
+		.libname = simple_basename(filename),
+	};
+
+	if (unlikely(mcount_should_stop()))
+		return ret;
+
+	mtdp = get_thread_data();
+	if (unlikely(check_thread_data(mtdp))) {
+		mtdp = mcount_prepare();
+		if (mtdp == NULL)
+			return ret;
+	}
+	else {
+		if (unlikely(mtdp->recursion_guard))
+			return ret;
+
+		mtdp->recursion_guard = true;
+	}
+
+	dl_iterate_phdr(dlopen_base_callback, &data);
+
+	/*
+	 * get timestamp before calling dlopen() so that
+	 * it can have symbols in static initializers which
+	 * called during the dlopen.
+	 */
+	send_dlopen_msg(mtdp, session_name(), timestamp,
+			data.base_addr, data.libname);
+
+	mtdp->recursion_guard = false;
+	return ret;
+}
+
+/*
+ * external interfaces
+ */
+void __visible_default __monstartup(unsigned long low, unsigned long high)
+{
+	mcount_startup();
+}
+
+void __visible_default _mcleanup(void)
+{
+	mcount_cleanup();
+}
+
+void __visible_default mcount_restore(void)
+{
+	mcount_rstack_restore();
 }
 
 void __visible_default mcount_reset(void)
 {
-	int idx;
-	struct mcount_thread_data *mtdp;
-
-	mtdp = get_thread_data();
-	if (unlikely(check_thread_data(mtdp)))
-		return;
-
-	for (idx = mtdp->idx - 1; idx >= 0; idx--)
-		*mtdp->rstack[idx].parent_loc = (unsigned long)mcount_return;
+	mcount_rstack_reset();
 }
 
 void __visible_default __cyg_profile_func_enter(void *child, void *parent)
@@ -876,11 +1153,11 @@ static void __attribute__((constructor))
 mcount_init(void)
 {
 	if (!mcount_setup_done)
-		__monstartup(0UL, ~0UL);
+		mcount_startup();
 }
 
 static void __attribute__((destructor))
 mcount_fini(void)
 {
-	_mcleanup();
+	mcount_cleanup();
 }
