@@ -20,10 +20,11 @@
 
 #define FTRACE_DIR_NAME   "uftrace.data"
 
-#define MCOUNT_RSTACK_MAX  1024
-#define MCOUNT_NOTRACE_IDX 0x10000
-#define MCOUNT_INVALID_DYNIDX 0xffff
-#define MCOUNT_DEFAULT_DEPTH  (INT_MAX - 1)
+#define MCOUNT_RSTACK_MAX      OPT_RSTACK_DEFAULT
+#define MCOUNT_DEFAULT_DEPTH   OPT_DEPTH_DEFAULT
+
+#define MCOUNT_NOTRACE_IDX     0x10000
+#define MCOUNT_INVALID_DYNIDX  0xffff
 
 enum mcount_flag {
 	MCOUNT_FL_SETJMP	= (1U << 0),
@@ -50,6 +51,7 @@ struct mcount_ret_stack {
 	uint64_t end_time;
 	int tid;
 	int filter_depth;
+	uint64_t filter_time;
 	unsigned short depth;
 	unsigned short dyn_idx;
 	/* set arg_spec at function entry and use it at exit */
@@ -91,6 +93,8 @@ struct filter_control {
 	int out_count;
 	int depth;
 	int saved_depth;
+	uint64_t time;
+	uint64_t saved_time;
 };
 #else
 struct filter_control {};
@@ -126,6 +130,7 @@ struct mcount_thread_data {
 	bool				recursion_guard;
 	bool				plthook_guard;
 	unsigned long			plthook_addr;
+	unsigned long			cygprof_dummy;
 	struct mcount_ret_stack		*rstack;
 	void				*argbuf;
 	struct filter_control		filter;
@@ -155,13 +160,16 @@ extern unsigned long plthook_resolver_addr;
 
 extern void __monstartup(unsigned long low, unsigned long high);
 extern void mcount_return(void);
-extern void mcount_prepare(void);
+extern struct mcount_thread_data * mcount_prepare(void);
 extern uint64_t mcount_gettime(void);
 extern bool mcount_check_rstack(struct mcount_thread_data *mtdp);
 extern void ftrace_send_message(int type, void *data, size_t len);
 extern const char *session_name(void);
 extern int gettid(struct mcount_thread_data *mtdp);
 
+extern void mcount_rstack_restore(void);
+extern void mcount_rstack_reset(void);
+
 extern void prepare_shmem_buffer(struct mcount_thread_data *mtdp);
 extern void get_new_shmem_buffer(struct mcount_thread_data *mtdp);
 extern void finish_shmem_buffer(struct mcount_thread_data *mtdp, int idx);
@@ -170,12 +178,13 @@ extern void shmem_finish(struct mcount_thread_data *mtdp);
 
 extern int hook_pltgot(char *exename, unsigned long offset);
 extern void plthook_setup(struct symtabs *symtabs);
+extern unsigned long plthook_return(void);
 extern void setup_dynsym_indexes(struct symtabs *symtabs);
 extern void destroy_dynsym_indexes(void);
 
 static inline bool mcount_should_stop(void)
 {
-	return !mcount_setup_done || mcount_finished || mtd.recursion_guard;
+	return !mcount_setup_done || mcount_finished;
 }
 
 struct ftrace_trigger;
