@@ -13,12 +13,12 @@ class TestCase(TestBase):
   backtrace [ 4629] | /* [ 1] alloc1 */
   backtrace [ 4629] | /* [ 2] alloc2 */
   backtrace [ 4629] | /* [ 3] alloc3 */
-            [ 4629] |         alloc4() {
-            [ 4629] |           alloc5() {
-   2.020 us [ 4629] |             malloc();
-   4.334 us [ 4629] |           } /* alloc5 */
-   4.671 us [ 4629] |         } /* alloc4 */
-""")
+            [ 4629] | alloc4() {
+            [ 4629] |   alloc5() {
+   2.020 us [ 4629] |     malloc();
+   4.334 us [ 4629] |   } /* alloc5 */
+   4.671 us [ 4629] | } /* alloc4 */
+""", sort='simple')
 
     def pre(self):
         record_cmd = '%s record -d %s %s' % (TestBase.ftrace, TDIR, 't-allocfree')
