@@ -26,7 +26,11 @@ class TestCase(TestBase):
     def pre(self):
         if os.geteuid() != 0:
             return TestBase.TEST_SKIP
-        record_cmd = '%s record -K3 -d %s %s' % (TestBase.ftrace, TDIR, 't-' + self.name)
+        if os.path.exists('/.dockerenv'):
+            return TestBase.TEST_SKIP
+
+        record_cmd = '%s record -K3 -N %s@kernel -d %s %s' % \
+                     (TestBase.ftrace, 'smp_irq_work_interrupt', TDIR, 't-' + self.name)
         sp.call(record_cmd.split())
         return TestBase.TEST_SUCCESS
 
