@@ -26,8 +26,11 @@ class TestCase(TestBase):
     def pre(self):
         if os.geteuid() != 0:
             return TestBase.TEST_SKIP
+        if os.path.exists('/.dockerenv'):
+            return TestBase.TEST_SKIP
+
         return TestBase.TEST_SUCCESS
 
     def runcmd(self):
-        return '%s -k --kernel-depth=2 -N %s@kernel %s' % \
-            (TestBase.ftrace, 'exit_to_usermode_loop', 't-' + self.name)
+        return '%s -k --kernel-depth=2 -N %s@kernel -N %s@kernel %s' % \
+            (TestBase.ftrace, 'exit_to_usermode_loop', 'smp_irq_work_interrupt', 't-' + self.name)
