@@ -63,6 +63,8 @@ static void print_trigger(struct ftrace_trigger *tr)
 
 	if (tr->flags & TRIGGER_FL_COLOR)
 		pr_dbg("\ttrigger: color '%c'\n", tr->color);
+	if (tr->flags & TRIGGER_FL_TIME_FILTER)
+		pr_dbg("\ttrigger: time filter %"PRIu64"\n", tr->time);
 }
 
 static bool match_ip(struct ftrace_filter *filter, unsigned long ip)
@@ -71,13 +73,13 @@ static bool match_ip(struct ftrace_filter *filter, unsigned long ip)
 }
 
 /**
- * ftrace_match_filter - try to match @ip with filters in @root
- * @root - root of rbtree which has filters
+ * uftrace_match_filter - try to match @ip with filters in @root
  * @ip   - instruction address to match
+ * @root - root of rbtree which has filters
  * @tr   - trigger data
  */
-struct ftrace_filter *ftrace_match_filter(struct rb_root *root, unsigned long ip,
-					  struct ftrace_trigger *tr)
+struct ftrace_filter *uftrace_match_filter(uint64_t ip, struct rb_root *root,
+					   struct ftrace_trigger *tr)
 {
 	struct rb_node *parent = NULL;
 	struct rb_node **p = &root->rb_node;
@@ -180,6 +182,8 @@ static void add_trigger(struct ftrace_filter *filter, struct ftrace_trigger *tr,
 
 	if (tr->flags & TRIGGER_FL_COLOR)
 		filter->trigger.color = tr->color;
+	if (tr->flags & TRIGGER_FL_TIME_FILTER)
+		filter->trigger.time = tr->time;
 }
 
 static void add_filter(struct rb_root *root, struct ftrace_filter *filter,
@@ -221,8 +225,7 @@ static void add_filter(struct rb_root *root, struct ftrace_filter *filter,
 }
 
 static int add_exact_filter(struct rb_root *root, struct symtab *symtab,
-			    char *module, char *filter_str,
-			    struct ftrace_trigger *tr)
+			    char *filter_str, struct ftrace_trigger *tr)
 {
 	struct ftrace_filter filter;
 	struct sym *sym;
@@ -240,8 +243,7 @@ static int add_exact_filter(struct rb_root *root, struct symtab *symtab,
 }
 
 static int add_regex_filter(struct rb_root *root, struct symtab *symtab,
-			    char *module, char *filter_str,
-			    struct ftrace_trigger *tr)
+			    char *filter_str, struct ftrace_trigger *tr)
 {
 	struct ftrace_filter filter;
 	struct sym *sym;
@@ -484,18 +486,15 @@ static int parse_float_argument_spec(char *str, struct ftrace_trigger *tr)
 	return 0;
 }
 
-static int setup_module_and_trigger(char *str, char *module,
-				    struct symtabs *symtabs,
+static int setup_module_and_trigger(char *str, struct symtabs *symtabs,
 				    struct symtab **psymtab,
-				    struct ftrace_trigger *tr)
+				    struct ftrace_trigger *tr,
+				    bool *found_mod)
 {
 	char *pos = strchr(str, '@');
 
-	*psymtab = &symtabs->symtab;
-
 	if (pos) {
 		char *tr_str;
-		bool found_mod = false;
 
 		*pos++ = '\0';
 		tr_str = xstrdup(pos);
@@ -580,13 +579,28 @@ static int setup_module_and_trigger(char *str, char *module,
 				continue;
 			}
 
-			if (module && strcasecmp(pos, module))
-				return -1;
+			if (!strncasecmp(pos, "time=", 5)) {
+				char *unit = NULL;
+
+				tr->flags |= TRIGGER_FL_TIME_FILTER;
+				tr->time = strtoull(pos+5, &unit, 10);
+
+				if (!strcmp(unit, "s"))
+					tr->time *= 1000 * 1000 * 1000;
+				else if (!strcmp(unit, "ms"))
+					tr->time *= 1000 * 1000;
+				else if (!strcmp(unit, "us"))
+					tr->time *= 1000;
+
+				continue;
+			}
 
 			if (!strcasecmp(pos, "plt"))
 				*psymtab = &symtabs->dsymtab;
 			else if (!strcasecmp(pos, "kernel"))
 				*psymtab = get_kernel_symtab();
+			else if (!strcmp(pos, basename(symtabs->filename)))
+				*psymtab = &symtabs->symtab;
 			else {
 				struct ftrace_proc_maps *map;
 
@@ -599,22 +613,15 @@ static int setup_module_and_trigger(char *str, char *module,
 				*psymtab = &map->symtab;
 			}
 
-			found_mod = true;
+			*found_mod = true;
 		}
-
-		if (module && !found_mod)
-			return -1;
-	}
-	else {
-		if (module)
-			return -1;
 	}
 
 	return 0;
 }
 
 static void setup_trigger(char *filter_str, struct symtabs *symtabs,
-			  char *module, struct rb_root *root,
+			  struct rb_root *root,
 			  unsigned long flags, enum filter_mode *fmode)
 {
 	char *str;
@@ -635,12 +642,13 @@ static void setup_trigger(char *filter_str, struct symtabs *symtabs,
 			.flags = flags,
 			.pargs = &args,
 		};
-		int ret;
-		char *mod = module;
+		int ret = 0;
+		bool mod_found = false;
 		struct ftrace_arg_spec *arg;
+		bool is_regex = strpbrk(name, REGEX_CHARS);
 
-		if (setup_module_and_trigger(name, mod, symtabs, &symtab,
-					     &tr) < 0)
+		if (setup_module_and_trigger(name, symtabs, &symtab,
+					     &tr, &mod_found) < 0)
 			goto next;
 
 		/* skip unintended kernel symbols */
@@ -654,20 +662,18 @@ static void setup_trigger(char *filter_str, struct symtabs *symtabs,
 			tr.fmode = FILTER_MODE_IN;
 
 again:
-		if (strpbrk(name, REGEX_CHARS))
-			ret = add_regex_filter(root, symtab, mod, name, &tr);
+		if (is_regex)
+			ret += add_regex_filter(root, symtab, name, &tr);
 		else
-			ret = add_exact_filter(root, symtab, mod, name, &tr);
+			ret += add_exact_filter(root, symtab, name, &tr);
 
-		if (ret == 0 && mod == NULL) {
-			mod = "plt";
+		if (!mod_found && (ret == 0 || is_regex)) {
 			symtab = &symtabs->dsymtab;
+			mod_found = true;
 			goto again;
 		}
 
-		/* kernel filter should not affect user functions */
-		if (ret > 0 && fmode != NULL &&
-		    (module == NULL || strcmp(module, "kernel"))) {
+		if (ret > 0 && fmode != NULL) {
 			if (tr.fmode == FILTER_MODE_IN)
 				*fmode = FILTER_MODE_IN;
 			else if (*fmode == FILTER_MODE_NONE)
@@ -691,57 +697,53 @@ next:
  * ftrace_setup_filter - construct rbtree of filters
  * @filter_str - CSV of filter string
  * @symtabs    - symbol tables to find symbol address
- * @module     - optional module (binary/dso) name
  * @root       - root of resulting rbtree
  * @mode       - filter mode: opt-in (-F) or opt-out (-N)
  */
 void ftrace_setup_filter(char *filter_str, struct symtabs *symtabs,
-			 char *module, struct rb_root *root,
-			 enum filter_mode *mode)
+			 struct rb_root *root, enum filter_mode *mode)
 {
-	setup_trigger(filter_str, symtabs, module, root, TRIGGER_FL_FILTER, mode);
+	setup_trigger(filter_str, symtabs, root, TRIGGER_FL_FILTER, mode);
 }
 
 /**
  * ftrace_setup_trigger - construct rbtree of triggers
  * @trigger_str - CSV of trigger string (FUNC @ act)
  * @symtabs    - symbol tables to find symbol address
- * @module     - optional module (binary/dso) name
  * @root       - root of resulting rbtree
  */
 void ftrace_setup_trigger(char *trigger_str, struct symtabs *symtabs,
-			  char *module, struct rb_root *root)
+			  struct rb_root *root)
 {
-	setup_trigger(trigger_str, symtabs, module, root, 0, NULL);
+	setup_trigger(trigger_str, symtabs, root, 0, NULL);
 }
 
 /**
  * ftrace_setup_argument - construct rbtree of argument
  * @args_str   - CSV of argument string (FUNC @ arg)
  * @symtabs    - symbol tables to find symbol address
- * @module     - optional module (binary/dso) name
  * @root       - root of resulting rbtree
  */
 void ftrace_setup_argument(char *args_str, struct symtabs *symtabs,
-			  char *module, struct rb_root *root)
+			   struct rb_root *root)
 {
-	setup_trigger(args_str, symtabs, module, root, 0, NULL);
+	setup_trigger(args_str, symtabs, root, 0, NULL);
 }
 
 /**
  * ftrace_setup_retval - construct rbtree of retval
  * @retval_str   - CSV of argument string (FUNC @ arg)
  * @symtabs    - symbol tables to find symbol address
- * @module     - optional module (binary/dso) name
  * @root       - root of resulting rbtree
  */
 void ftrace_setup_retval(char *retval_str, struct symtabs *symtabs,
-			  char *module, struct rb_root *root)
+			 struct rb_root *root)
 {
-	setup_trigger(retval_str, symtabs, module, root, 0, NULL);
+	setup_trigger(retval_str, symtabs, root, 0, NULL);
 }
 
-void ftrace_setup_filter_module(char *trigger_str, struct list_head *head)
+void ftrace_setup_filter_module(char *trigger_str, struct list_head *head,
+				const char *modname)
 {
 	char *str, *tmp;
 	char *pos, *name, *action;
@@ -787,6 +789,9 @@ void ftrace_setup_filter_module(char *trigger_str, struct list_head *head)
 					continue;
 			}
 
+			if (!strcmp(pos, basename(modname)))
+				goto next;
+
 			list_for_each_entry(fm, head, list) {
 				if (!strcasecmp(fm->name, pos))
 					goto next;
@@ -894,7 +899,7 @@ TEST_CASE(filter_setup_exact)
 	filter_test_load_symtabs(&stabs);
 
 	/* test1: simple method */
-	ftrace_setup_filter("foo::bar", &stabs, NULL, &root, NULL);
+	ftrace_setup_filter("foo::bar", &stabs, &root, NULL);
 	TEST_EQ(RB_EMPTY_ROOT(&root), false);
 
 	node = rb_first(&root);
@@ -907,7 +912,7 @@ TEST_CASE(filter_setup_exact)
 	TEST_EQ(RB_EMPTY_ROOT(&root), true);
 
 	/* test2: destructor */
-	ftrace_setup_filter("foo::~foo", &stabs, NULL, &root, NULL);
+	ftrace_setup_filter("foo::~foo", &stabs, &root, NULL);
 	TEST_EQ(RB_EMPTY_ROOT(&root), false);
 
 	node = rb_first(&root);
@@ -920,7 +925,7 @@ TEST_CASE(filter_setup_exact)
 	TEST_EQ(RB_EMPTY_ROOT(&root), true);
 
 	/* test3: unknown symbol */
-	ftrace_setup_filter("invalid_name", &stabs, NULL, &root, NULL);
+	ftrace_setup_filter("invalid_name", &stabs, &root, NULL);
 	TEST_EQ(RB_EMPTY_ROOT(&root), true);
 
 	return TEST_OK;
@@ -937,7 +942,7 @@ TEST_CASE(filter_setup_regex)
 
 	filter_test_load_symtabs(&stabs);
 
-	ftrace_setup_filter("foo::b.*", &stabs, NULL, &root, NULL);
+	ftrace_setup_filter("foo::b.*", &stabs, &root, NULL);
 	TEST_EQ(RB_EMPTY_ROOT(&root), false);
 
 	node = rb_first(&root);
@@ -982,11 +987,11 @@ TEST_CASE(filter_setup_notrace)
 
 	filter_test_load_symtabs(&stabs);
 
-	ftrace_setup_filter("foo::.*", &stabs, NULL, &root, &fmode);
+	ftrace_setup_filter("foo::.*", &stabs, &root, &fmode);
 	TEST_EQ(RB_EMPTY_ROOT(&root), false);
 	TEST_EQ(fmode, FILTER_MODE_IN);
 
-	ftrace_setup_filter("!foo::foo", &stabs, NULL, &root, &fmode);
+	ftrace_setup_filter("!foo::foo", &stabs, &root, &fmode);
 	TEST_EQ(RB_EMPTY_ROOT(&root), false);
 	TEST_EQ(fmode, FILTER_MODE_IN);  /* overall filter mode doesn't change */
 
@@ -1021,26 +1026,26 @@ TEST_CASE(filter_match)
 
 	filter_test_load_symtabs(&stabs);
 
-	ftrace_setup_filter("foo::foo", &stabs, NULL, &root, &fmode);
+	ftrace_setup_filter("foo::foo", &stabs, &root, &fmode);
 	TEST_EQ(RB_EMPTY_ROOT(&root), false);
 	TEST_EQ(fmode, FILTER_MODE_IN);
 
 	memset(&tr, 0, sizeof(tr));
-	TEST_NE(ftrace_match_filter(&root, 0x1000, &tr), NULL);
+	TEST_NE(uftrace_match_filter(0x1000, &root, &tr), NULL);
 	TEST_EQ(tr.flags, TRIGGER_FL_FILTER);
 	TEST_EQ(tr.fmode, FILTER_MODE_IN);
 
 	memset(&tr, 0, sizeof(tr));
-	TEST_NE(ftrace_match_filter(&root, 0x1fff, &tr), NULL);
+	TEST_NE(uftrace_match_filter(0x1fff, &root, &tr), NULL);
 	TEST_EQ(tr.flags, TRIGGER_FL_FILTER);
 	TEST_EQ(tr.fmode, FILTER_MODE_IN);
 
 	memset(&tr, 0, sizeof(tr));
-	TEST_EQ(ftrace_match_filter(&root, 0xfff, &tr), NULL);
+	TEST_EQ(uftrace_match_filter(0xfff, &root, &tr), NULL);
 	TEST_NE(tr.flags, TRIGGER_FL_FILTER);
 
 	memset(&tr, 0, sizeof(tr));
-	TEST_EQ(ftrace_match_filter(&root, 0x2000, &tr), NULL);
+	TEST_EQ(uftrace_match_filter(0x2000, &root, &tr), NULL);
 	TEST_NE(tr.flags, TRIGGER_FL_FILTER);
 
 	ftrace_cleanup_filter(&root);
@@ -1061,27 +1066,27 @@ TEST_CASE(trigger_setup)
 
 	filter_test_load_symtabs(&stabs);
 
-	ftrace_setup_trigger("foo::bar@depth=2", &stabs, NULL, &root);
+	ftrace_setup_trigger("foo::bar@depth=2", &stabs, &root);
 	TEST_EQ(RB_EMPTY_ROOT(&root), false);
 
 	memset(&tr, 0, sizeof(tr));
-	TEST_NE(ftrace_match_filter(&root, 0x2500, &tr), NULL);
+	TEST_NE(uftrace_match_filter(0x2500, &root, &tr), NULL);
 	TEST_EQ(tr.flags, TRIGGER_FL_DEPTH);
 	TEST_EQ(tr.depth, 2);
 
-	ftrace_setup_trigger("foo::bar@backtrace", &stabs, NULL, &root);
+	ftrace_setup_trigger("foo::bar@backtrace", &stabs, &root);
 	memset(&tr, 0, sizeof(tr));
-	TEST_NE(ftrace_match_filter(&root, 0x2500, &tr), NULL);
+	TEST_NE(uftrace_match_filter(0x2500, &root, &tr), NULL);
 	TEST_EQ(tr.flags, TRIGGER_FL_DEPTH | TRIGGER_FL_BACKTRACE);
 
-	ftrace_setup_trigger("foo::baz1@traceon", &stabs, NULL, &root);
+	ftrace_setup_trigger("foo::baz1@traceon", &stabs, &root);
 	memset(&tr, 0, sizeof(tr));
-	TEST_NE(ftrace_match_filter(&root, 0x3000, &tr), NULL);
+	TEST_NE(uftrace_match_filter(0x3000, &root, &tr), NULL);
 	TEST_EQ(tr.flags, TRIGGER_FL_TRACE_ON);
 
-	ftrace_setup_trigger("foo::baz3@trace_off,depth=1", &stabs, NULL, &root);
+	ftrace_setup_trigger("foo::baz3@trace_off,depth=1", &stabs, &root);
 	memset(&tr, 0, sizeof(tr));
-	TEST_NE(ftrace_match_filter(&root, 0x5000, &tr), NULL);
+	TEST_NE(uftrace_match_filter(0x5000, &root, &tr), NULL);
 	TEST_EQ(tr.flags, TRIGGER_FL_TRACE_OFF | TRIGGER_FL_DEPTH);
 	TEST_EQ(tr.depth, 1);
 
