@@ -32,6 +32,9 @@
 
 static bool kernel_tracing_enabled;
 
+static int save_kernel_files(struct ftrace_kernel *kernel);
+static int load_kernel_files(struct ftrace_kernel *kernel);
+
 
 static char *get_tracing_file(const char *name)
 {
@@ -505,15 +508,192 @@ int finish_kernel_tracing(struct ftrace_kernel *kernel)
 	free(kernel->traces);
 	free(kernel->fds);
 
+	save_kernel_files(kernel);
+	save_kernel_symbol(kernel->output_dir);
+
 	reset_tracing_files();
 
 	return 0;
 }
 
+static const char *get_endian_str(void)
+{
+	if (get_elf_endian() == ELFDATA2LSB)
+		return "LE";
+	else
+		return "BE";
+}
+
+static int save_kernel_file(FILE *fp, const char *name)
+{
+	int fd;
+	ssize_t len;
+	char buf[4096];
+
+	snprintf(buf, sizeof(buf), "%s/%s", TRACING_DIR, name);
+
+	fd = open(buf, O_RDONLY);
+	if (fd < 0)
+		return -1;
+
+	len = read(fd, buf, sizeof(buf));
+	if (len < 0) {
+		close(fd);
+		return -1;
+	}
+
+	fprintf(fp, "TRACEFS: %s: %zd\n", name, len);
+	fwrite(buf, len, 1, fp);
+
+	close(fd);
+	return 0;
+}
+
+static int save_kernel_files(struct ftrace_kernel *kernel)
+{
+	char *path = NULL;
+	FILE *fp;
+	int ret = -1;
+
+	xasprintf(&path, "%s/kernel_header", kernel->output_dir);
+
+	fp = fopen(path, "w");
+	if (fp == NULL)
+		pr_err("cannot write kernel header");
+
+	fprintf(fp, "PAGE_SIZE: %d\n", getpagesize());
+	fprintf(fp, "LONG_SIZE: %zd\n", sizeof(long));
+	fprintf(fp, "ENDIAN: %s\n", get_endian_str());
+
+	if (save_kernel_file(fp, "events/header_page") < 0)
+		goto out;
+
+	if (save_kernel_file(fp, "events/ftrace/funcgraph_entry/format") < 0)
+		goto out;
+
+	if (save_kernel_file(fp, "events/ftrace/funcgraph_exit/format") < 0)
+		goto out;
+
+	ret = 0;
+
+out:
+	fclose(fp);
+	free(path);
+	return ret;
+}
+
+/* provided for backward compatibility */
+static int load_current_kernel(struct ftrace_kernel *kernel)
+{
+	int fd;
+	size_t len;
+	char buf[4096];
+	bool is_big_endian = !strcmp(get_endian_str(), "BE");
+	struct pevent *pevent = kernel->pevent;
+
+	pevent_set_long_size(pevent, sizeof(long));
+	pevent_set_page_size(pevent, getpagesize());
+	pevent_set_file_bigendian(pevent, is_big_endian);
+	pevent_set_host_bigendian(pevent, is_big_endian);
+
+	fd = open(TRACING_DIR"/events/header_page", O_RDONLY);
+	if (fd < 0)
+		return -1;
+
+	len = read(fd, buf, sizeof(buf));
+	pevent_parse_header_page(pevent, buf, len, sizeof(long));
+	close(fd);
+
+	fd = open(TRACING_DIR"/events/ftrace/funcgraph_entry/format", O_RDONLY);
+	if (fd < 0)
+		return -1;
+
+	len = read(fd, buf, sizeof(buf));
+	pevent_parse_event(pevent, buf, len, "ftrace");
+	close(fd);
+
+	fd = open(TRACING_DIR"/events/ftrace/funcgraph_exit/format", O_RDONLY);
+	if (fd < 0)
+		return -1;
+
+	len = read(fd, buf, sizeof(buf));
+	pevent_parse_event(pevent, buf, len, "ftrace");
+	close(fd);
+
+	return 0;
+}
+
+static int load_kernel_files(struct ftrace_kernel *kernel)
+{
+	char *path = NULL;
+	FILE *fp;
+	char buf[4096];
+	struct pevent *pevent = kernel->pevent;
+	int ret = 0;
+
+	xasprintf(&path, "%s/kernel_header", kernel->output_dir);
+
+	fp = fopen(path, "r");
+	if (fp == NULL)  /* old data doesn't have the kernel header */
+		return load_current_kernel(kernel);
+
+	while (fgets(buf, sizeof(buf), fp) != NULL) {
+		char name[128];
+		size_t len;
+
+		if (strncmp(buf, "TRACEFS:", 8) != 0) {
+			char val[32];
+
+			sscanf(buf, "%[^:]: %s\n", name, val);
+
+			if (!strcmp(name, "PAGE_SIZE")) {
+				ret = strtol(val, NULL, 0);
+				pevent_set_page_size(pevent, ret);
+			}
+			else if (!strcmp(name, "LONG_SIZE")) {
+				ret = strtol(val, NULL, 0);
+				pevent_set_long_size(pevent, ret);
+			}
+			else if (!strcmp(name, "ENDIAN")) {
+				ret = (strcmp(val, "BE") == 0);
+				pevent_set_file_bigendian(pevent, ret);
+
+				ret = (strcmp(get_endian_str(), "BE") == 0);
+				pevent_set_host_bigendian(pevent, ret);
+			}
+			continue;
+		}
+
+		sscanf(buf, "TRACEFS: %[^:]: %zd\n", name, &len);
+
+		if (fread(buf, 1, len, fp) != len) {
+			ret = -1;
+			break;
+		}
+
+		if (!strcmp(name, "events/header_page")) {
+			pevent_parse_header_page(pevent, buf, len,
+						 pevent_get_long_size(pevent));
+		}
+		else if (!strcmp(name, "events/ftrace/funcgraph_entry/format") ||
+			 !strcmp(name, "events/ftrace/funcgraph_exit/format")) {
+			ret = pevent_parse_event(pevent, buf, len, "ftrace");
+			if (ret != 0) {
+				pevent_strerror(pevent, ret, buf, len);
+				pr_err_ns("%s: %s\n", name, buf);
+			}
+		}
+	}
+
+	fclose(fp);
+	free(path);
+	return ret;
+}
+
 static size_t trace_pagesize;
 static struct trace_seq trace_seq;
 static struct ftrace_ret_stack trace_rstack = {
-	.unused = FTRACE_UNUSED,
+	.magic = RECORD_MAGIC,
 };
 
 static int prepare_kbuffer(struct ftrace_kernel *kernel, int cpu);
@@ -541,8 +721,6 @@ static int scandir_filter(const struct dirent *d)
 int setup_kernel_data(struct ftrace_kernel *kernel)
 {
 	int i;
-	int fd;
-	size_t len;
 	char buf[4096];
 	enum kbuffer_endian endian = KBUFFER_ENDIAN_LITTLE;
 	enum kbuffer_long_size longsize = KBUFFER_LSIZE_8;
@@ -560,6 +738,11 @@ int setup_kernel_data(struct ftrace_kernel *kernel)
 		return -1;
 	}
 
+	if (load_kernel_files(kernel) < 0) {
+		pr_out("cannot read kernel header: %m\n");
+		return -1;
+	}
+
 	pr_dbg("found kernel ftrace data for %d cpus\n", kernel->nr_cpus);
 
 	kernel->fds	= xcalloc(kernel->nr_cpus, sizeof(*kernel->fds));
@@ -575,12 +758,11 @@ int setup_kernel_data(struct ftrace_kernel *kernel)
 	kernel->missed_events = xcalloc(kernel->nr_cpus, sizeof(*kernel->missed_events));
 	kernel->tids          = xcalloc(kernel->nr_cpus, sizeof(*kernel->tids));
 
-	/* FIXME: should read recorded data file */
 	if (pevent_is_file_bigendian(kernel->pevent))
 		endian = KBUFFER_ENDIAN_BIG;
 	if (pevent_get_long_size(kernel->pevent) == 4)
 		longsize = KBUFFER_LSIZE_4;
-	trace_pagesize = getpagesize(); /* pevent_get_page_size() */
+	trace_pagesize = pevent_get_page_size(kernel->pevent);
 
 	for (i = 0; i < kernel->nr_cpus; i++) {
 		struct stat stbuf;
@@ -617,32 +799,6 @@ int setup_kernel_data(struct ftrace_kernel *kernel)
 		return -1;
 	}
 
-	fd = open(TRACING_DIR"/events/header_page", O_RDONLY);
-	if (fd < 0)
-		return -1;
-
-	len = read(fd, buf, sizeof(buf));
-	pevent_parse_header_page(kernel->pevent, buf, len, pevent_get_long_size(kernel->pevent));
-	close(fd);
-
-	fd = open(TRACING_DIR"/events/ftrace/funcgraph_entry/format", O_RDONLY);
-	if (fd < 0)
-		return -1;
-
-	len = read(fd, buf, sizeof(buf));
-	pevent_parse_event(kernel->pevent, buf, len, "ftrace");
-	close(fd);
-
-	fd = open(TRACING_DIR"/events/ftrace/funcgraph_exit/format", O_RDONLY);
-	if (fd < 0)
-		return -1;
-
-	len = read(fd, buf, sizeof(buf));
-	pevent_parse_event(kernel->pevent, buf, len, "ftrace");
-	close(fd);
-
-	/* TODO: read /proc/kallsyms and register functions */
-
 	pevent_register_event_handler(kernel->pevent, -1, "ftrace", "funcgraph_entry",
 				      funcgraph_entry_handler, NULL);
 	pevent_register_event_handler(kernel->pevent, -1, "ftrace", "funcgraph_exit",
@@ -696,7 +852,8 @@ static int prepare_kbuffer(struct ftrace_kernel *kernel, int cpu)
 	kernel->mmaps[cpu] = mmap(NULL, trace_pagesize, PROT_READ, MAP_PRIVATE,
 				  kernel->fds[cpu], kernel->offsets[cpu]);
 	if (kernel->mmaps[cpu] == MAP_FAILED) {
-		pr_dbg("loading kbuffer for cpu %d failed", cpu);
+		pr_dbg("loading kbuffer for cpu %d (fd: %d, offset: %lu, pagesize: %zd) failed\n",
+		       cpu, kernel->fds[cpu], kernel->offsets[cpu], trace_pagesize);
 		return -1;
 	}
 
@@ -824,9 +981,6 @@ static int read_kernel_cpu(struct ftrace_file_handle *handle, int cpu)
 	struct ftrace_ret_stack *curr;
 	int tid, prev_tid = -1;
 
-	if (!handle->time_filter)
-		return read_kernel_cpu_data(kernel, cpu);
-
 	if (rstack_list->count)
 		goto out;
 
@@ -835,30 +989,75 @@ static int read_kernel_cpu(struct ftrace_file_handle *handle, int cpu)
 	 * the given time filter (-t option).
 	 */
 	while (read_kernel_cpu_data(kernel, cpu) == 0) {
+		struct ftrace_session *sess = first_session;
+		struct ftrace_task_handle *task;
+		struct ftrace_trigger tr = {};
+		uint64_t real_addr;
+		uint64_t time_filter = handle->time_filter;
+
 		curr = &kernel->rstacks[cpu];
 
 		/* prevent ustack from invalid access */
 		kernel->rstack_valid[cpu] = false;
 
 		tid = kernel->tids[cpu];
+		task = get_task_handle(handle, tid);
+		if (task == NULL)
+			continue;
+
+		if (!check_time_range(&handle->time_range, curr->time))
+			continue;
+
 		if (prev_tid == -1)
 			prev_tid = tid;
 
-		/* XXX: handle scheduled task properly */
-		if (tid != prev_tid) {
-			add_to_rstack_list(rstack_list, curr, NULL);
-			break;
-		}
-		prev_tid = tid;
+		if (task->filter.time)
+			time_filter = task->filter.time->threshold;
+
+		/* filter match needs full (64-bit) address */
+		real_addr = get_real_address(curr->addr);
+		/*
+		 * it might set TRACE trigger, which shows
+		 * function even if it's less than the time filter.
+		 */
+		uftrace_match_filter(real_addr, &sess->filters, &tr);
 
 		if (curr->type == FTRACE_ENTRY) {
 			/* it needs to wait until matching exit found */
 			add_to_rstack_list(rstack_list, curr, NULL);
+
+			if (tr.flags & TRIGGER_FL_TIME_FILTER) {
+				struct time_filter_stack *tfs;
+
+				tfs = xmalloc(sizeof(*tfs));
+				tfs->next = task->filter.time;
+				tfs->depth = curr->depth;
+				tfs->context = FSTACK_CTX_KERNEL;
+				tfs->threshold = tr.time;
+
+				task->filter.time = tfs;
+			}
+
+			/* XXX: handle scheduled task properly */
+			if (tid != prev_tid)
+				break;
 		}
 		else if (curr->type == FTRACE_EXIT) {
 			struct uftrace_rstack_list_node *last;
 			uint64_t delta;
 
+			if (task->filter.time) {
+				struct time_filter_stack *tfs;
+
+				tfs = task->filter.time;
+				if (tfs->depth == curr->depth &&
+				    tfs->context == FSTACK_CTX_KERNEL) {
+					/* discard stale filter */
+					task->filter.time = tfs->next;
+					free(tfs);
+				}
+			}
+
 			if (rstack_list->count == 0) {
 				/* it's already exceeded time filter, just return */
 				add_to_rstack_list(rstack_list, curr, NULL);
@@ -869,19 +1068,7 @@ static int read_kernel_cpu(struct ftrace_file_handle *handle, int cpu)
 					       typeof(*last), list);
 			delta = curr->time - last->rstack.time;
 
-			if (delta < handle->time_filter) {
-				struct ftrace_session *sess = first_session;
-				struct ftrace_trigger tr = {};
-				unsigned long real_addr;
-
-				/* filter match needs full (64-bit) address */
-				real_addr = get_real_address(curr->addr);
-				/*
-				 * it might set TRACE trigger, which shows
-				 * function even if it's less than the time filter.
-				 */
-				ftrace_match_filter(&sess->filters,
-						    real_addr, &tr);
+			if (delta < time_filter) {
 				if (tr.flags & TRIGGER_FL_TRACE) {
 					add_to_rstack_list(rstack_list, curr, NULL);
 					break;
@@ -889,6 +1076,10 @@ static int read_kernel_cpu(struct ftrace_file_handle *handle, int cpu)
 
 				/* also delete matching entry (at the last) */
 				delete_last_rstack_list(rstack_list);
+
+				/* XXX: handle scheduled task properly */
+				if (tid != prev_tid)
+					break;
 			} else {
 				/* found! process all existing rstacks in the list */
 				add_to_rstack_list(rstack_list, curr, NULL);
@@ -901,6 +1092,7 @@ static int read_kernel_cpu(struct ftrace_file_handle *handle, int cpu)
 			break;
 		}
 
+		prev_tid = tid;
 	}
 	if (kernel->rstack_done[cpu] && rstack_list->count == 0)
 		return -1;
@@ -935,8 +1127,10 @@ int read_kernel_stack(struct ftrace_file_handle *handle,
 	int first_tid = -1;
 	uint64_t first_timestamp = 0;
 	struct ftrace_kernel *kernel = handle->kern;
-	struct ftrace_ret_stack *first_rstack = NULL;
+	struct ftrace_ret_stack *first_rstack;
 
+retry:
+	first_rstack = NULL;
 	for (i = 0; i < kernel->nr_cpus; i++) {
 		uint64_t timestamp;
 
@@ -962,6 +1156,12 @@ int read_kernel_stack(struct ftrace_file_handle *handle,
 		return -1;
 
 	*taskp = get_task_handle(handle, first_tid);
+	if (*taskp == NULL) {
+		/* force re-read on that cpu */
+		kernel->rstack_valid[first_cpu] = false;
+		goto retry;
+	}
+
 	memcpy(&(*taskp)->kstack, first_rstack, sizeof(*first_rstack));
 
 	return first_cpu;
